from openfisca_us.model_api import *


class ny_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "NY taxable income"
    unit = USD
    documentation = "NY AGI less taxable income deductions"
    definition_period = YEAR



from openfisca_us.model_api import *


class md_non_single_childless_refundable_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "MD refundable EITC for filers who are not single and childless"
    unit = USD
    definition_period = YEAR
    reference = "https://casetext.com/statute/code-of-maryland/article-tax-general/title-10-income-tax/subtitle-7-income-tax-credits/section-10-704-effective-until-6302023-for-earned-income"
    defined_for = StateCode.MD

    def formula(tax_unit, period, parameters):
        single_childless = tax_unit(
            "md_qualifies_for_single_childless_eitc", period
        )
        in_md = tax_unit.household("state_code_str", period) == "MD"
        # Must have zeroed out MD tax liability with non-refundable EITC.
        md_tax_before_credits = tax_unit(
            "md_income_tax_before_credits", period
        )
        md_non_single_childless_non_refundable_eitc = tax_unit(
            "md_non_single_childless_non_refundable_eitc", period
        )
        md_tax_equals_non_refundable_eitc = (
            md_tax_before_credits
            == md_non_single_childless_non_refundable_eitc
        )
        eligible = (
            in_md & ~single_childless & md_tax_equals_non_refundable_eitc
        )
        federal_eitc_without_age_minimum = tax_unit(
            "federal_eitc_without_age_minimum", period
        )
        p = parameters(period).gov.states.md.tax.income.credits.eitc
        matched_eitc = p.refundable_match * federal_eitc_without_age_minimum
        return eligible * max_(0, matched_eitc - md_tax_before_credits)



from openfisca_us.model_api import *


class md_dependent_care_subtraction(Variable):
    value_type = float
    entity = TaxUnit
    label = "MD depdendent care subtraction from AGI"
    unit = USD
    definition_period = YEAR
    reference = "https://casetext.com/statute/code-of-maryland/article-tax-general/title-10-income-tax/subtitle-2-maryland-taxable-income-calculations-for-individual/part-ii-maryland-adjusted-gross-income/section-10-208-effective-until-712024-subtractions-from-federal-adjusted-gross-income-state-adjustments"
    defined_for = StateCode.MD

    """
    (e) The subtraction under subsection (a) of this section includes expenses
    for household and dependent care services not exceeding the dollar limit
    allowed under § 21(c) of the Internal Revenue Code and determined without
    reference to the percentage limitation in § 21(a)(2) of the Internal Revenue Code.

    IRC § 21(c): https://www.law.cornell.edu/uscode/text/26/21#c
    IRC § 21(a)(2): https://www.law.cornell.edu/uscode/text/26/21#a_2
    """

    formula = sum_of_variables(["cdcc_relevant_expenses"])



from openfisca_us.model_api import *


class aged_blind_extra_standard_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Aged and blind standard deduction"
    definition_period = YEAR
    unit = USD

    def formula(tax_unit, period, parameters):
        std = parameters(period).gov.irs.deductions.standard
        filing_status = tax_unit("filing_status", period)
        filing_status_type = filing_status.possible_values
        blind_head = tax_unit("blind_head", period) * 1
        blind_spouse = tax_unit("blind_spouse", period) * 1
        aged_head = (
            tax_unit("age_head", period) >= std.aged_or_blind.age_threshold
        ) * 1
        aged_spouse = (
            (filing_status == filing_status_type.JOINT)
            & (
                tax_unit("age_spouse", period)
                >= std.aged_or_blind.age_threshold
            )
        ) * 1
        count_extra_stded = blind_head + blind_spouse + aged_head + aged_spouse
        return count_extra_stded * std.aged_or_blind.amount[filing_status]



from openfisca_us.model_api import *


class tax_exempt_social_security(Variable):
    value_type = float
    entity = TaxUnit
    label = "Tax-exempt Social Security"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        total_ss = tax_unit("tax_unit_social_security", period)
        taxable_ss = tax_unit("tax_unit_taxable_social_security", period)
        return total_ss - taxable_ss



from openfisca_us.model_api import *


class mo_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "MO taxable income"
    unit = USD
    definition_period = YEAR
    reference = "https://revisor.mo.gov/main/OneSection.aspx?section=143.111&bid=7201&hl="
    defined_for = StateCode.MO



from openfisca_us.model_api import *


class snap_child_support_deduction(Variable):
    value_type = float
    entity = SPMUnit
    label = "SNAP child support payment deduction"
    unit = USD
    documentation = (
        "Deduction from SNAP gross income for child support payments"
    )
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/7/2014#e_4"

    def formula(spm_unit, period, parameters):
        child_support = add(spm_unit, period, ["child_support_expense"])
        state = spm_unit.household("state_code_str", period)
        is_deductible = parameters(
            period
        ).gov.usda.snap.income.deductions.child_support[state]
        return where(is_deductible, child_support, 0)



from openfisca_us.model_api import *


class in_other_taxes(Variable):
    value_type = float
    entity = TaxUnit
    label = "IN other taxes"
    unit = USD
    definition_period = YEAR
    documentation = "Includes the use tax on out-of-state purchases, household employment taxes, and recapture of certain Indiana offset credits."



from openfisca_us.model_api import *


class in_homeowners_property_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "IN rent"
    unit = USD
    definition_period = YEAR
    documentation = "Property taxes paid on a principal place of residence that was subject to Indiana property tax."
    reference = (
        "http://iga.in.gov/legislative/laws/2021/ic/titles/006#6-3-1-3.5"
    )



from openfisca_us.model_api import *


class md_single_childless_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "MD single childless EITC"
    unit = USD
    definition_period = YEAR
    reference = "https://casetext.com/statute/code-of-maryland/article-tax-general/title-10-income-tax/subtitle-7-income-tax-credits/section-10-704-effective-until-6302023-for-earned-income"  # (c)(3)
    defined_for = StateCode.MD

    def formula(tax_unit, period, parameters):
        single_childless = tax_unit(
            "md_qualifies_for_single_childless_eitc", period
        )
        in_md = tax_unit.household("state_code_str", period) == "MD"
        eligible = single_childless & in_md
        federal_eitc_without_age_minimum = tax_unit(
            "federal_eitc_without_age_minimum", period
        )
        p = parameters(period).gov.states.md.tax.income.credits.eitc.childless
        match = p.percent_match * federal_eitc_without_age_minimum
        cap = p.max_amount
        return eligible * min_(match, cap)



from openfisca_us.model_api import *


class wa_income_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "Washington income tax"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.WA

    def formula(tax_unit, period, parameters):
        in_wa = tax_unit.household("state_code_str", period) == "WA"
        wftc = tax_unit("wa_working_families_tax_credit", period)
        cap_gains = tax_unit("wa_capital_gains_tax", period)
        tax = cap_gains - wftc
        return tax * in_wa



from openfisca_us.model_api import *


class StateName(Enum):
    # Source: https://gist.github.com/rogerallen/1583593#gistcomment-3699885
    # 50 states and DC.
    AL = "Alabama"
    AK = "Alaska"
    AZ = "Arizona"
    AR = "Arkansas"
    CA = "California"
    CO = "Colorado"
    CT = "Connecticut"
    DE = "Delaware"
    FL = "Florida"
    GA = "Georgia"
    HI = "Hawaii"
    ID = "Idaho"
    IL = "Illinois"
    IN = "Indiana"
    IA = "Iowa"
    KS = "Kansas"
    KY = "Kentucky"
    LA = "Louisiana"
    ME = "Maine"
    MD = "Maryland"
    MA = "Massachusetts"
    MI = "Michigan"
    MN = "Minnesota"
    MS = "Mississippi"
    MO = "Missouri"
    MT = "Montana"
    NE = "Nebraska"
    NV = "Nevada"
    NH = "New Hampshire"
    NJ = "New Jersey"
    NM = "New Mexico"
    NY = "New York"
    NC = "North Carolina"
    ND = "North Dakota"
    OH = "Ohio"
    OK = "Oklahoma"
    OR = "Oregon"
    PA = "Pennsylvania"
    RI = "Rhode Island"
    SC = "South Carolina"
    SD = "South Dakota"
    TN = "Tennessee"
    TX = "Texas"
    UT = "Utah"
    VT = "Vermont"
    VA = "Virginia"
    WA = "Washington"
    WV = "West Virginia"
    WI = "Wisconsin"
    WY = "Wyoming"
    DC = "District of Columbia"
    # Territories.
    GU = "Guam"
    MP = "Northern Mariana Islands"
    PW = "Palau"
    PR = "Puerto Rico"
    VI = "Virgin Islands"
    AA = "Armed Forces Americas (Except Canada)"
    AE = "Armed Forces Africa/Canada/Europe/Middle East"
    AP = "Armed Forces Pacific"


class state_name(Variable):
    value_type = Enum
    possible_values = StateName
    default_value = StateName.CA
    entity = Household
    label = "State"
    definition_period = ETERNITY

    def formula(household, period, parameters):
        fips = household("fips", period)
        return (
            pd.Series(fips)
            .map(
                {
                    1: StateName.AL,
                    2: StateName.AK,
                    4: StateName.AZ,
                    5: StateName.AR,
                    6: StateName.CA,
                    8: StateName.CO,
                    9: StateName.CT,
                    10: StateName.DE,
                    11: StateName.DC,
                    12: StateName.FL,
                    13: StateName.GA,
                    15: StateName.HI,
                    16: StateName.ID,
                    17: StateName.IL,
                    18: StateName.IN,
                    19: StateName.IA,
                    20: StateName.KS,
                    21: StateName.KY,
                    22: StateName.LA,
                    23: StateName.ME,
                    24: StateName.MD,
                    25: StateName.MA,
                    26: StateName.MI,
                    27: StateName.MN,
                    28: StateName.MS,
                    29: StateName.MO,
                    30: StateName.MT,
                    31: StateName.NE,
                    32: StateName.NV,
                    33: StateName.NH,
                    34: StateName.NJ,
                    35: StateName.NM,
                    36: StateName.NY,
                    37: StateName.NC,
                    38: StateName.ND,
                    39: StateName.OH,
                    40: StateName.OK,
                    41: StateName.OR,
                    42: StateName.PA,
                    44: StateName.RI,
                    45: StateName.SC,
                    46: StateName.SD,
                    47: StateName.TN,
                    48: StateName.TX,
                    49: StateName.UT,
                    50: StateName.VT,
                    51: StateName.VA,
                    53: StateName.WA,
                    54: StateName.WV,
                    55: StateName.WI,
                    56: StateName.WY,
                    72: StateName.PR,
                }
            )
            .values
        )



from openfisca_us.model_api import *


class meets_ssi_resource_test(Variable):
    value_type = bool
    entity = Person
    label = "Meets SSI resource test"
    unit = USD
    definition_period = YEAR

    def formula(person, period, parameters):
        abd = person("is_ssi_aged_blind_disabled", period)
        joint_claim = person("ssi_claim_is_joint", period)
        personal_resources = person("ssi_countable_resources", period)
        countable_resources = where(
            joint_claim,
            person.marital_unit.sum(personal_resources),
            personal_resources,
        )
        ssi = parameters(period).gov.ssa.ssi
        resource_limits = ssi.eligibility.resources.limit
        resource_limit = where(
            joint_claim, resource_limits.couple, resource_limits.individual
        )
        return countable_resources <= resource_limit



from openfisca_us.model_api import *


class c05800(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Income tax before credits"
    unit = USD
    documentation = "Total (regular + AMT) income tax liability before credits"

    formula = sum_of_variables(
        [
            "income_tax_main_rates",
            "capital_gains_tax",
            "alternative_minimum_tax",
        ]
    )


income_tax_before_credits = variable_alias("income_tax_before_credits", c05800)



from openfisca_us.model_api import *


class net_long_term_capital_gain(Variable):
    value_type = float
    entity = TaxUnit
    label = "Net long-term capital gain"
    unit = USD
    documentation = (
        "The excess of long-term capital gains over long-term capital losses."
    )
    definition_period = YEAR
    reference = dict(
        title="26 U.S. Code § 1222(7)",
        href="https://www.law.cornell.edu/uscode/text/26/1222#7",
    )

    formula = excess(
        of="long_term_capital_gains", over="long_term_capital_losses"
    )



from openfisca_us.model_api import *
from openfisca_core.tracers import SimpleTracer


class tax_liability_if_not_itemizing(Variable):
    value_type = float
    entity = TaxUnit
    label = "Tax liability if not itemizing"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        simulation = tax_unit.simulation
        simulation.max_spiral_loops = 10
        simulation._check_for_cycle = lambda *args: None
        simulation_if_not_itemizing = simulation.clone()
        computed_variables = get_stored_variables(simulation)
        simulation_if_not_itemizing.tracer = simulation.tracer
        simulation_if_not_itemizing.set_input(
            "tax_unit_itemizes",
            period,
            np.zeros((tax_unit.count,), dtype=bool),
        )
        values = simulation_if_not_itemizing.calculate(
            "federal_state_income_tax", period
        )
        added_variables = set(
            get_stored_variables(simulation_if_not_itemizing)
        ) - set(computed_variables)
        for variable in added_variables:
            simulation.get_holder(variable).delete_arrays()
        return values



from openfisca_us.model_api import *


class utility_expense(Variable):
    value_type = float
    entity = SPMUnit
    label = "Utility expenses"
    unit = USD
    definition_period = YEAR

    formula = sum_of_variables(
        [
            "heating_cooling_expense",
            "gas_expense",
            "electricity_expense",
            "trash_expense",
            "water_expense",
            "sewage_expense",
            "phone_expense",
        ]
    )



from openfisca_us.model_api import *


class household_net_income(Variable):
    value_type = float
    entity = Household
    label = "Household net income"
    definition_period = YEAR
    unit = USD

    def formula(household, period, parameters):
        return household.sum(household.members("net_income", period))



from openfisca_us.model_api import *


class income_tax_capped_non_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Non-refundable tax credits"
    unit = USD
    documentation = "Capped value of non-refundable tax credits"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return min_(
            tax_unit("income_tax_before_credits", period),
            tax_unit("income_tax_non_refundable_credits", period),
        )



from openfisca_tools import IndividualSim as GeneralIndividualSim
from openfisca_us import CountryTaxBenefitSystem
from openfisca_us.entities import entities
from openfisca_us.data import CPS


class IndividualSim(GeneralIndividualSim):
    tax_benefit_system = CountryTaxBenefitSystem
    entities = {entity.key: entity for entity in entities}
    default_dataset = CPS

    default_roles = dict(
        tax_unit="member",
        spm_unit="member",
        household="member",
        family="member",
    )
    required_entities = [
        "tax_unit",
        "spm_unit",
        "household",
        "family",
    ]



from openfisca_us.model_api import *


class self_employment_tax_ald(Variable):
    value_type = float
    entity = TaxUnit
    label = "Self-employment tax ALD deduction"
    unit = USD
    documentation = "Above-the-line deduction for self-employment tax"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/164#f"

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        self_employment_tax = person("self_employment_tax", period)
        is_dependent = person("is_tax_unit_dependent", period)
        total_tax = tax_unit.sum(self_employment_tax * ~is_dependent)
        percent_deductible = parameters(
            period
        ).gov.irs.ald.self_employment_tax.percent_deductible
        return total_tax * percent_deductible



from openfisca_us.model_api import *


class is_ctc_qualifying_child(Variable):
    value_type = bool
    entity = Person
    label = "Is a CTC-qualifying child"
    documentation = "Is a child qualifying for the Child Tax Credit"
    definition_period = YEAR

    def formula(person, period, parameters):
        age = person("age", period)
        return age <= parameters(period).gov.irs.credits.ctc.child.max_age



from openfisca_us.model_api import *


class state_income_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "State income tax"
    unit = USD
    definition_period = YEAR

    formula_2021 = sum_of_variables(
        ["ma_income_tax", "wa_income_tax", "md_income_tax"]
    )



from openfisca_us.model_api import *


class in_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "IN deductions"
    unit = USD
    definition_period = YEAR
    reference = "http://iga.in.gov/legislative/laws/2021/ic/titles/006#6-3"

    formula = sum_of_variables(
        [
            "in_renters_deduction",
            "in_homeowners_property_tax_deduction",
            "salt_refund_last_year",
            "us_govt_interest",
            "tax_unit_taxable_social_security",  # includes railroad retirement benefits
            "in_military_service_deduction",
            "in_nonpublic_school_deduction",
            "in_nol",
            "in_unemployment_compensation_deduction",
            "in_other_deductions",
        ]
    )



from openfisca_us.model_api import *


class mo_property_tax_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "MO property tax credit"
    unit = USD
    definition_period = YEAR
    reference = ()
    defined_for = StateCode.MO



from openfisca_us.model_api import *


class md_non_single_childless_non_refundable_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = (
        "MD non-refundable EITC for filers who are not single and childless"
    )
    unit = USD
    definition_period = YEAR
    reference = "https://casetext.com/statute/code-of-maryland/article-tax-general/title-10-income-tax/subtitle-7-income-tax-credits/section-10-704-effective-until-6302023-for-earned-income"  # (c)(1)
    defined_for = StateCode.MD

    def formula(tax_unit, period, parameters):
        federal_eitc = tax_unit("eitc", period)
        p = parameters(period).gov.states.md.tax.income.credits.eitc
        tax_before_credits = tax_unit("md_income_tax_before_credits", period)
        # Limited to filers who are not single and childless.
        single_childless = tax_unit(
            "md_qualifies_for_single_childless_eitc", period
        )
        in_md = tax_unit.household("state_code_str", period) == "MD"
        eligible = ~single_childless & in_md
        uncapped = p.non_refundable_match * federal_eitc
        return eligible * min_(tax_before_credits, uncapped)



from openfisca_us.model_api import *


class ca_is_qualifying_child_for_caleitc(Variable):
    value_type = bool
    entity = Person
    label = "Child qualifies for CalEITC"
    definition_period = YEAR
    reference = "https://www.ftb.ca.gov/file/personal/credits/california-earned-income-tax-credit.html#What-you-ll-get"



from openfisca_us.model_api import *


class snap_gross_income(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = YEAR
    label = "SNAP gross income"
    documentation = "Gross income for calculating SNAP eligibility"
    reference = "https://www.law.cornell.edu/uscode/text/7/2014#d"
    unit = USD

    def formula(spm_unit, period):
        ELEMENTS = ["snap_earned_income", "snap_unearned_income"]
        return add(spm_unit, period, ELEMENTS)



from openfisca_us.model_api import *


class md_non_refundable_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "MD EITC non-refundable State tax credit"
    unit = USD
    documentation = "Non-refundable EITC credit reducing MD State income tax."
    definition_period = YEAR
    reference = "https://casetext.com/statute/code-of-maryland/article-tax-general/title-10-income-tax/subtitle-7-income-tax-credits/section-10-704-effective-until-6302023-for-earned-income"
    defined_for = StateCode.MD

    def formula(tax_unit, period, parameters):
        federal_eitc = tax_unit("eitc", period)
        md_eitc = parameters(period).gov.states.md.tax.income.credits.eitc
        childless = tax_unit("eitc_child_count", period) == 0
        return where(
            childless,
            min_(md_eitc.childless_max, federal_eitc),
            md_eitc.non_refundable_match * federal_eitc,
        )



from openfisca_us.model_api import *


class qualified_business_income(Variable):
    value_type = float
    entity = Person
    label = "Qualified business income"
    documentation = "Income connected with the trade or conduct of a qualified trade or business. A qualified trade or business is any trade or business other than a specified service trade or business, or employment. The list of specified service trades can be found at https://www.law.cornell.edu/uscode/text/26/1202#e_3_A."
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/199A#c"

    def formula(tax_unit, period, parameters):
        components = parameters(
            period
        ).gov.irs.deductions.qbi.max.income_definition
        total_income = add(tax_unit, period, components)
        qualified = tax_unit("business_is_qualified", period)
        return total_income * qualified

    # Note. does not implement https://www.law.cornell.edu/uscode/text/26/199A#d_3, which provides exceptions for where some specified service business income may be
    # treated as qualified business income depending on taxable income.



from matplotlib import units
import yaml
from openfisca_us.model_api import *
from openfisca_core.taxbenefitsystems import TaxBenefitSystem


def import_yaml():
    import yaml

    try:
        from yaml import CLoader as Loader
    except ImportError:
        from yaml import SafeLoader as Loader
    return yaml, Loader


yaml, Loader = import_yaml()


def create_taxcalc_alias(name: str, variable: Type[Variable]):
    """Creates a new OpenFisca variable with the same metadata as the given variable, but the Tax-Calculator name.
    The variable will be cast to tax unit level if it is not already.

    Args:
        name (str): The name of the equivalent Tax-Calculator variable.
        variable (Type[Variable]): The OpenFisca-US variable class.
    """

    full_name = "taxcalc_" + name

    def formula(tax_unit, period, parameters):
        if variable.entity == TaxUnit:
            return tax_unit(variable.__name__, period)
        elif variable.entity == Person:
            return add(tax_unit, period, [variable.__name__])
        elif variable.entity == Household:
            return tax_unit.value_from_first_person(
                tax_unit.members.household(variable.__name__, period)
            )
        else:
            raise ValueError("Unsupported entity: {}".format(variable.entity))

    original_documentation = (
        variable.documentation if hasattr(variable, "documentation") else None
    )
    addition_to_documentation = (
        "This is a read-only variable alias returning the OpenFisca-US variable "
        + name
        + " under the (prefixed) name for its equivalent in Tax-Calculator, "
        + full_name
        + "."
    )
    documentation = (
        original_documentation + ". " + addition_to_documentation
        if original_documentation
        else addition_to_documentation
    )

    return type(
        full_name,
        (Variable,),
        dict(
            entity=TaxUnit,
            definition_period=YEAR,
            label=variable.label + " (Tax-Calculator)"
            if hasattr(variable, "label")
            else name + "(Tax-Calculator)",
            unit=variable.unit if hasattr(variable, "unit") else None,
            documentation=documentation,
            value_type=variable.value_type,
            formula=formula,
        ),
    )


def add_taxcalc_variable_aliases(system: TaxBenefitSystem):
    """Adds aliases for all Tax-Calculator variables to the given TaxBenefitSystem.

    Args:
        system (TaxBenefitSystem): The TaxBenefitSystem to which the aliases should be added.
    """

    with open(Path(__file__).parent / "variable_mapping.yaml") as f:
        variable_map = yaml.load(f, Loader=Loader)

    for openfisca_us_name, taxcalc_name in variable_map.items():
        try:
            system.add_variable(
                create_taxcalc_alias(
                    taxcalc_name, type(system.variables[openfisca_us_name])
                )
            )
        except Exception as e:
            print("Error adding alias for {}: {}".format(openfisca_us_name, e))
            raise e



from openfisca_us.model_api import *


class employee_medicare_tax(Variable):
    value_type = float
    entity = Person
    label = "Employee-side health insurance payroll tax"
    definition_period = YEAR
    unit = USD

    def formula(person, period, parameters):
        rate = parameters(period).gov.irs.payroll.medicare.rate.employee
        return rate * person("payroll_tax_gross_wages", period)



from openfisca_us.model_api import *


class ut_additions_to_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "UT additions to income"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.UT



from openfisca_us.model_api import *


class qualified_dividend_income(Variable):
    value_type = float
    entity = Person
    label = "Qualified dividend income"
    unit = USD
    definition_period = YEAR

    def formula(person, period, parameters):
        qualified_percentage = parameters(
            period
        ).calibration.programs.income.dividends.qualified_percentage
        total_dividends = person("dividend_income", period)
        return total_dividends * qualified_percentage



from openfisca_us.model_api import *


class health_insurance_premiums(Variable):
    value_type = float
    entity = Person
    label = "Health insurance premiums"
    unit = USD
    definition_period = YEAR



from openfisca_us.model_api import *


class bi_before_phase_out(Variable):
    value_type = float
    entity = Person
    label = "Basic income before phase-outs"
    unit = USD
    documentation = (
        "Total basic income payments for this person, before apply phase-outs."
    )
    definition_period = YEAR

    def formula(person, period, parameters):
        bi = parameters(period).contrib.ubi_center.basic_income
        age = person("age", period)
        return bi.amount_by_age.calc(age)



from openfisca_us.model_api import *


class non_qualified_dividend_income(Variable):
    value_type = float
    entity = Person
    label = "Ordinary dividend income"
    unit = USD
    definition_period = YEAR

    def formula(person, period, parameters):
        total_dividends = person("dividend_income", period)
        qualified_dividends = person("qualified_dividend_income", period)
        return total_dividends - qualified_dividends



from openfisca_us.model_api import *
import numpy as np


class ca_young_child(Variable):
    value_type = float
    entity = TaxUnit
    label = "CA Young Child Tax Credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.ftb.ca.gov/forms/2021/2021-3514.pdf#page=3",
        "https://leginfo.legislature.ca.gov/faces/codes_displaySection.xhtml?lawCode=RTC&sectionNum=17052.1",
        "https://www.ftb.ca.gov/forms/2021/2021-3514-instructions.html",
    )

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ca.tax.income.credits.young_child
        earnings = tax_unit("tax_unit_earned_income", period)

        # Eligibility requires
        # a) tax unit receives CalEITC
        # b) tax unit has at least one CalEITC-qualifying child under six
        person = tax_unit.members
        meets_age_limit = person("age", period) < p.ineligible_age
        is_qualifying_child_for_caleitc = person(
            "ca_is_qualifying_child_for_caleitc", period
        )
        is_eligible_child = meets_age_limit & is_qualifying_child_for_caleitc
        has_eligible_child = tax_unit.sum(is_eligible_child) > 0
        receives_ca_eitc = tax_unit("ca_eitc", period) > 0
        in_ca = tax_unit.household("state_code_str", period) == "CA"
        eligible = has_eligible_child & receives_ca_eitc & in_ca
        # Phase out the amount.
        # NB: Law and instructions say $20 per $100, but the tax form
        # instructs the filer to round.
        excess_earnings = max_(0, earnings - p.phase_out.start)
        # Round increments to two decimal places.
        increments = np.round_(excess_earnings / p.phase_out.increment, 2)
        # Round reduction to two decimal places.
        reduction = min_(
            p.amount, np.round_(increments * p.phase_out.amount, 2)
        )
        # Round final result to nearest dollar.
        # In reality, <$1 goes to $1.
        return eligible * np.rint(p.amount - reduction)



from openfisca_us.model_api import *


class income_tax_before_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    unit = USD
    label = "Income tax before refundable credits"
    documentation = "Income tax liability (including othertaxes) after non-refundable credits are used, but before refundable credits are applied"

    def formula(tax_unit, period, parameters):
        income_tax_bc = tax_unit("income_tax_before_credits", period)
        capped_credits = tax_unit(
            "income_tax_capped_non_refundable_credits", period
        )
        taxes_net_nonrefundable_credits = income_tax_bc - capped_credits
        OTHER_TAXES = [
            "net_investment_income_tax",
            "recapture_of_investment_credit",
            "unreported_payroll_tax",
            "qualified_retirement_penalty",
        ]
        other_taxes = add(tax_unit, period, OTHER_TAXES)
        return taxes_net_nonrefundable_credits + other_taxes



from openfisca_us.data import CPS
import pytest
from openfisca_us import Microsimulation

CPS_YEARS = [2020]


@pytest.mark.dependency(name="cps")
@pytest.mark.parametrize("year", CPS_YEARS)
def test_cps_dataset_generates(year):
    CPS.generate(year)


@pytest.mark.dependency(depends=["cps"])
@pytest.mark.parametrize("year", CPS_YEARS)
def test_cps_openfisca_us_compatible(year):
    Microsimulation(dataset=CPS, year=year).calc("employment_income")



from openfisca_us.model_api import *


class is_ctc_qualifying_young_child(Variable):
    value_type = bool
    entity = Person
    label = "Is a CTC-qualifying young child"
    documentation = (
        "Is a child qualifying for the Child Tax Credit young child bonus"
    )
    definition_period = YEAR

    def formula(person, period, parameters):
        age = person("age", period)
        return (
            age <= parameters(period).gov.irs.credits.ctc.child.young.max_age
        )
